
import asyncio

from .clu_client import CluClient
from .exceptions import FeatureNotGettableError, FeatureNotSettableError
from .interfaces import FeatureInterface
from .types import DataType


class GFeature:
    _clu_client: CluClient
    _object_id: str
    _name: str
    _index: int
    _settable: bool
    _gettable: bool
    _data_type: DataType
    _unit: str

    _enum: dict | None
    _value_range: tuple[int, int] | None
    
    def __init__(self, clu_client: CluClient, object_id: str, interface: FeatureInterface) -> None:
        self._clu_client = clu_client
        self._object_id = object_id
        self._name = interface.name
        self._index = interface.index
        self._settable = interface.set
        self._gettable = interface.get
        self._data_type = interface.data_type
        self._unit = interface.unit
        self._enum = interface.enum
        self._value_range = interface.value_range

    @property
    def name(self) -> str:
        return self._name

    @property
    def parent(self) -> str:
        return self._object_id

    @property
    def index(self) -> int:
        return self._index

    @property
    def is_settable(self) -> bool:
        return self._settable

    @property
    def is_gettable(self) -> bool:
        return self._gettable

    @property
    def data_type(self) -> DataType:
        return self._data_type

    @property
    def unit(self) -> str:
        return self._unit

    @property
    def enum(self) -> dict | None:
        return self._enum

    @property
    def value_range(self) -> tuple[int, int] | None:
        return self._value_range

    async def get_value_async(self):
        if not self._gettable:
            raise FeatureNotGettableError(self._name)

        value = await self._clu_client.get_value_async(self._object_id, self._index)
        
        return self.data_type.convert_value(value)

    def get_value(self):
        return asyncio.get_event_loop().run_until_complete(self.get_value_async())

    async def get_value_mapped_async(self):
        val = await self.get_value_async()
        
        if self._enum is None or val not in self._enum.keys():
            return val
        
        return self._enum[val]
    
    def get_value_mapped(self):
        return asyncio.get_event_loop().run_until_complete(self.get_value_mapped_async())
    
    async def set_value_async(self, value):
        if not self._settable:
            raise FeatureNotSettableError(self._name)

        if self._enum is not None and value not in self._enum.keys():
            raise ValueError(f"Value: {value} is not in enum: {self._enum}")
        if self._value_range is not None and (value < self._value_range[0] or value > self._value_range[1]):
            raise ValueError(f"Value: {value} is not in value range: ({self._value_range[0]} - {self._value_range[1]})")
    
        await self._clu_client.set_value_async(self._object_id, self._index, value)

    def set_value(self, value) -> None:
        asyncio.get_event_loop().run_until_complete(self.set_value_async(value))
        
    def register_handler(self, handler) -> None:
        self._clu_client.register_value_change_handler(self._object_id, self._index, handler)

    def remove_handler(self) -> None:
        self._clu_client.remove_value_change_handler(self._object_id, self._index)




import logging
from dataclasses import dataclass

from ..clu_client import CluClient
from ..config_objects import CLUConfig
from ..gobject import GObject
from ..interface_manager import InterfaceManager
from ..interfaces import CluObjectInterface, ModuleObjectInterface
from ..types import ModuleObjectType
from .om_parser import OMEndpoints


@dataclass
class CluConfig():
    om_config: OMEndpoints
    conf_json: CLUConfig
    
    objects_by_class: dict[int, list[GObject]]
    objects_by_id: dict[str, GObject]
    objects_by_name: dict[str, GObject]    

def parse_clu_config(conf_json: CLUConfig, om: OMEndpoints, interface_manager: InterfaceManager, clu_client: CluClient) -> CluConfig:
    clu_interface = interface_manager.get_clu_interface(conf_json.hw_type, conf_json.fw_type, conf_json.fw_api_version)

    objects_by_class: dict[int, list[GObject]] = {}
    objects_by_id: dict[str, GObject] = {}
    objects_by_name: dict[str, GObject] = {}
    
    def add_object(obj: GObject) -> None:
        if obj.object_class in objects_by_class.keys():
            objects_by_class[obj.object_class].append(obj)
        else:
            objects_by_class[obj.object_class] = [obj]
        
        objects_by_id[obj.object_id] = obj
        objects_by_name[obj.name] = obj
    
    gclu = GObject(clu_client, om.getName(om.this_clu.object_id), om.this_clu.object_id, clu_interface)
    add_object(gclu)

    for sn, mod_objects in om.module_objects.items():
        mod_conf = conf_json.modules[sn]
        mod_int = interface_manager.get_module_interface(mod_conf.hw_type, mod_conf.fw_type, mod_conf.fw_api_version)
        
        if mod_int is None:
            logging.warn("Missing module interface. hw_type: %d, fw_api_version: %d.", mod_conf.hw_type, mod_conf.fw_api_version)
            
            for obj in mod_objects:
                empty_interface = ModuleObjectInterface("unknown", obj.object_class, ModuleObjectType.NONE, [], [])
                gobj = GObject(clu_client, om.getName(obj.object_id), obj.object_id, empty_interface)
                add_object(gobj) 
            continue
        
        for obj in mod_objects:
            obj_int = mod_int.objects[obj.object_class]
            obj_name = om.getName(obj.object_id)
            
            gobj = GObject(clu_client, obj_name, obj.object_id, obj_int)
            add_object(gobj)
        
    for clu_obj in om.clu_objects:
        if clu_obj.object_class not in clu_interface.objects.keys():
            logging.warn("Missing clu objectc interface. object_type: %d.", clu_obj.object_class)
            
            empty_interface = CluObjectInterface("unknown", clu_obj.object_class, 0)
            gobj = GObject(clu_client, om.getName(clu_obj.object_id), clu_obj.object_id, empty_interface)
            add_object(gobj)
            continue
        
        obj_int = clu_interface.objects[clu_obj.object_class]
        obj_name = om.getName(clu_obj.object_id)
        
        gclu_obj = GObject(clu_client, obj_name, clu_obj.object_id, obj_int)
        add_object(gclu_obj)

    return CluConfig(om, conf_json, objects_by_class, objects_by_id, objects_by_name)



DIN = 3
DOUT = 4
DIMM = 5
TIMER = 6
BUTTON = 7
LEDRGB = 9
LEDRGBW = 9
GSM = 10
ROLLER_SH = 11
ANALOGIN = 12
POWERSUPPLYVOLTAGE = 12
POWER_SUPPLY_VOLTAGE = 12
ANALOGOUT = 13
PIDCONTROLLER = 14
CALENDAR = 15
SCHEDULER = 16
GENERIC_ZW_NODE = 17
THERMOSTAT = 18
PANEL_PAGE = 19
TEMPERATURE_SENSOR = 20
PANELSENSTEMP = 20
PANELSENSLIGHT = 21
HUMIDITY_SENSOR = 22
ONEW_SENSOR = 23
ROLLER_SHUTTER = 24
WAKE_UP = 24
ZWAVE_BINARY_SENSOR = 25
BINARY_SENSOR = 25
DOUT_L = 26
DALI_MASTER = 27
PANEL = 29
PANEL_BUTTON = 30
PUSH = 31
DALI_GEAR = 32
PRESSURE_SENSOR = 34
DALI_GEAR_DT8 = 35
AIR_VOC_SENSOR = 36
SOUND_SENSOR = 37
IR_CONTROLLER = 38
SUNRISESUNSETCALENDAR = 39
PRESENCESENSOR = 40
AIR_CO2_SENSOR = 41
LIGHT_SENSOR_LUX = 42
EVENTSCHEDULER = 43
ZWAVE_THERMOSTAT = 51
ZWAVE_DOUT = 52
ZWAVE_DIN = 53
ZWAVE_WAKEUP = 54
ZWAVE_BATTERY = 55
ZWAVE_TAMPER_ALARM = 56
ZWAVE_POWER_METER = 57
ZWAVE_ANALOG_SENSOR = 58
ZWAVE_ROLLER_SHUTTER = 59
ZWAVE_DIMMER = 60
ZWAVE_UNKNOWN_MODULE = 61
ZWAVE_RGBW_LED = 62
ZWAVE_1W_SENSOR = 63
ZWAVE_CONFIG = 64
ZWAVE_FAKRO = 65
ZWAVE_FAKRO_ZWS = 65
ZWAVE_IR = 66
ZWAVE_IR_EP = 67
ZWAVE_WEATHER = 68
SATELZONE = 100
ZONE = 100
INPUT = 101
SATELINPUT = 101
OUTPUT = 102
SATELOUTPUT = 102
SATEL = 103
HTTPLISTENER = 120
HTTPREQUEST = 121
JABLOTRON = 150
JABLOTRONSECTION = 151
JABLOTRONOUTPUT = 152
JABLOTRONDEVICE = 153
MODBUS = 200
MODBUSVALUE = 201
COOLMASTERNET = 250
COOLMASTER = 252
SONOS = 260
MUSICCAST = 261




from datetime import datetime
from enum import Enum


class CallType(Enum):
    EXECUTE = "execute"
    GET = "get"
    SET = "set"

class DataType(Enum):
    #TODO: make data types more generic
    VOID = "void"
    NUMBER = "num"
    INTEGER = "int"
    ENUM = "enum"
    STRING = "str"
    TABLE = "table"
    TIMESTAMP = "timestamp"
    CONFIRMATION = "confirmation"
    NONE = ""
    
    def convert_value(self, var):
        if self.value == DataType.NUMBER:
            return float(var)
        elif self.value == DataType.INTEGER:
            return int(var)
        elif self.value == DataType.TIMESTAMP:
            return datetime.fromtimestamp(float(var))
        else:
            return var
class ModuleObjectType(Enum):
    NONE = "none"
    OUT = "out"
    IN = "in"



import re
import ipaddress
from ipaddress import IPv4Address
from dataclasses import dataclass, field

_NAME_FIELD = re.compile("-- NAME_.*")
_SPLIT_ARGS = re.compile("[ =,\n]+")
_OBJECT_CREATION = re.compile('.*[^"]OBJECT:new\(.*?\)[^"\n]*')

@dataclass
class CLU:
    object_id: str
    ipaddress: IPv4Address

@dataclass
class CLUObject:
    object_id: str
    object_class: int
    
@dataclass
class ModuleObject(CLUObject):
    parent_serial_number: int
    object_index: int

@dataclass
class OMEndpoints:
    this_clu: CLU
    external_clus: list[CLU] = field(default_factory=list)
    
    names: dict[str, str] = field(default_factory=dict)
    
    clu_objects: list[CLUObject] = field(default_factory=list)
    module_objects: dict[int, list[ModuleObject]] = field(default_factory=list)
    
    def getName(self, object_id) -> str:
        if object_id not in self.names.keys():
            return f"UNNAMED_OBJECT_{object_id}"
        
        return self.names[object_id]

def _parse_name(line: str) -> tuple[str, str]:
    spl = re.split(_SPLIT_ARGS, line)[2:]

    name = spl[0]
    object_id = spl[1]

    return object_id, name

def _parse_object_creation(line: str) -> tuple[str, list[str]]:
    obj_id = line.split()[0]
    args = line[line.find("(")+1:line.rfind(")")].split(", ")
    
    return obj_id, args

def _parse_ip_hex(hex_str: str) -> IPv4Address:
    return ipaddress.ip_address(int(hex_str, 16))

def parse_om(file) -> OMEndpoints:
    names: dict[str, str] = {}
    module_objects: dict[int, list[ModuleObject]] = {}
    clu_objects: list[CLUObject] = []
    
    this_clu: CLU = None
    external_clus: list[CLU] = []
            
    for line in file:
        if re.match(_NAME_FIELD, line):
            obj_id, name = _parse_name(line)
            names[obj_id] = name

        elif re.match(_OBJECT_CREATION, line):
            obj_id, args = _parse_object_creation(line)
            argcount = len(args)
            
            if argcount == 2:
                # parse clu object
                clu_objects.append(CLUObject(obj_id, int(args[0])))

            elif argcount == 3 :
                # parse this clu or external clu
                ip_address = _parse_ip_hex(args[1])
                
                if int(args[0]) == 1:
                    external_clus.append(CLU(obj_id, ip_address))
                else:
                    this_clu = CLU(obj_id, ip_address)
                
            elif argcount == 4:
                # parse module object
                obj_class = int(args[0])
                if obj_class == 2:
                    continue
                
                parent_serial_number = int(args[1][3:])
                index = int(args[2])
                
                if parent_serial_number in module_objects.keys():
                    module_objects[parent_serial_number].append(ModuleObject(obj_id, obj_class, parent_serial_number, index))
                else:
                    module_objects[parent_serial_number] = [ModuleObject(obj_id, obj_class, parent_serial_number, index)]

    return OMEndpoints(this_clu, external_clus, names, clu_objects, module_objects)
           



import io
import logging
import pickle
import re
from zipfile import ZipFile

import requests

from .interfaces import CluInterface, ModuleInterface
from .parsers.interfaces_parser import parse_interfaces

_OM_INTERFACES_ENDPOINT = "http://om.grenton.com/interfaces/v4/"
_OM_NEWEST_INTERFACES = "device-interfaces.current"

def _get_newest_verion() -> str:
    resp = requests.get(_OM_INTERFACES_ENDPOINT + _OM_NEWEST_INTERFACES)
    result = re.search(r"device.*zip", resp.text)
    if result is None:
        logging.error("Unable to get latest version of interface")
        return None
    return result.group(0)

def _get_current_version(dir) -> str | None:
    try:
        version_file = open(dir + "/interfaces-version.txt", "r")
        return version_file.readline()
    except IOError:
        return None
    
def _download_interfaces(version, directory) -> None:
    resp = requests.get(_OM_INTERFACES_ENDPOINT + version)
    
    if len(resp.content) == 0:
        logging.error("Unable to download interfaces.")
        return
    
    try:
        with ZipFile(io.BytesIO(resp.content)) as zip:
            zip.extractall(directory)
            
        with open(directory + "/interfaces-version.txt", "w") as version_file:
            version_file.write(version)
    except IOError as e:
        logging.error("Unable to save interfaces. Error message: %s", e.strerror)
    
class InterfaceManager:
    #TODO: maybe create database for intefaces
    _clus: dict[int, list[CluInterface]] = None
    _modules: dict[int, list[ModuleInterface]] = None

    def __init__(self, cache_dir: str) -> None:
        self._dir = cache_dir

        current_version = _get_current_version(cache_dir)
        newest_version = _get_newest_verion()
        
        if current_version != newest_version and newest_version is not None:
            _download_interfaces(newest_version, cache_dir)
            self._parse_interfaces()
            self._save_interface_cache()
        else:
            try:
                self._load_interface_cache()
            except:
                self._parse_interfaces()

    def get_clu_interface(self, hw_type: int, fw_type: int, api_version: int) -> CluInterface | None:
        clu_list = self._clus[hw_type]

        if clu_list is None:
            return None
        
        for clu in clu_list:
            if clu.fw_api_version == api_version and clu.fw_type == fw_type:
                return clu
            
        return clu_list[-1]

    def get_module_interface(self, hw_type: int, fw_type: int, api_version: int) -> ModuleInterface | None:
        mod_list = self._modules[hw_type]

        if mod_list is None:
            return None
        
        for mod in mod_list:
            if mod.fw_api_version == api_version and mod.fw_type == fw_type:
                return mod
            
        return mod_list[-1]
    
    def _parse_interfaces(self) -> None:
        clus, modules = parse_interfaces(self._dir + "/device-interfaces")
        self._clus = clus
        self._modules = modules
        self._save_interface_cache()
    
    def _save_list(self, list, filename) -> None:
        with open(self._dir + '/' + filename, 'wb') as file:
            pickle.dump(list, file)
    
    def _save_interface_cache(self) -> None:
        self._save_list(self._clus, "clus.cache")
        self._save_list(self._modules, "modules.cache")
        
    def _load_list(self, filename) -> list:
        with open(self._dir + "/" + filename, "rb") as file:
            return pickle.load(file)
        
    def _load_interface_cache(self) -> None:
        self._clus = self._load_list("clus.cache")
        self._modules = self._load_list("modules.cache")
    



import socket
import random

def get_host_ip(clu_ip: str) -> str:
    _, _, ips = socket.gethostbyname_ex(socket.gethostname())

    for ip in ips:
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.bind((ip, 0))
        
        try:
            sock.sendto(b"", (clu_ip, 1234))
            return ip
        except:
            pass
        
    #TODO: throw exception when no ip is found
    return ""

def find_n_character(string: str, char: str, n: int) -> int:
    count = 0
    for index, c in enumerate(string):
        if c == char:
            count += 1
            if count == n:
                return index
            
    return -1

def extract_payload(resp: str) -> str:
    index = find_n_character(resp, ':', 3)
    if index == -1:
        return resp
         
    return resp[index + 1:]
        
def generate_id_hex(lenght=8) -> str:
    return ''.join(random.choices("01234567890abcdef", k=lenght))

def parse_list(msg: str, start: int = 0) -> tuple[int, list]:
    values = []
    i = start
    prev_i = start
    quote = False
    
    def convert_type(data: str):
        data = data.strip()
        if data.startswith('"') and data.endswith('"'):
            return data[1:-1]
        elif data == "true":
            return True
        elif data == "false":
            return False
        elif data == "nil":
            return None
        
        return float(data)
    
    def append_if_not_empty(m: str):
        if m.strip() != "":
            values.append(convert_type(m))
    
    length = len(msg)
    while i < length:
        c = msg[i]
        if c == ',' and not quote:
            append_if_not_empty(msg[prev_i:i])
            prev_i = i+1 
        elif c == '"':
            quote = not quote
        elif c == '{' and not quote:
            i, sub_list = parse_list(msg, i+1)
            prev_i = i + 1
            values.append(sub_list)   
        elif c == '}' and not quote:
            append_if_not_empty(msg[prev_i:i])
            return i, values
        
        i += 1
    
    values.append(convert_type(msg[prev_i:]))
    return i-1, values


import asyncio
import socket
import threading
import time
from dataclasses import dataclass, field
from collections.abc import Callable
from typing import Any

from .cipher import GrentonCipher     
from .utils import (
    find_n_character,
    parse_list,
    get_host_ip,
    generate_id_hex,
    extract_payload
)

def _parse_update_message(msg: str) -> tuple[int, list]:
    index = find_n_character(msg, ':', 4)
    msg = msg[index + 1:]
    
    index = msg.find(':')
    client_id = int(msg[:index])
    
    msg = msg[index + 2:-1]
    _, values = parse_list(msg)
    
    return client_id, values

@dataclass
class UpdateContext:
    object_id: str
    index: int
    value: Any
    
@dataclass(eq=True, unsafe_hash=True)
class FeatureEntry:
    object_id: str
    index: int
    
@dataclass
class ClientPage:
    client_id: int
    features: list[FeatureEntry] = field(default_factory=list)
    modified: bool = False
    last_mod: float = field(default_factory=time.time)
    states: list[Any] = field(default_factory=list)
    
    def create_payload(self, client_ip, client_port) -> str:
        features_str = '{' + ','.join([f"{{{fe.object_id},{fe.index}}}" for fe in self.features]) + '}'
        return f'SYSTEM:clientRegister("{client_ip}",{client_port},{self.client_id},{features_str})'
    
    def __hash__(self) -> int:
        return hash(self.client_id)

class CluClient:

    def __init__(
        self,
        ip: str,
        port: int,
        cipher: GrentonCipher,
        timeout: float = 1,
        client_refresh_interval: float = 60,
        client_ip: str | None = None,
        client_port: int = 0,
        max_connections: int = 4,
        page_size: int = 16
    ) -> None:
        self._addr = (ip, port)
        self._timeout = timeout
        self._client_refresh_interval = client_refresh_interval
        
        if client_ip:
            self._local_ip = client_ip
        else:
            self._local_ip = get_host_ip(ip)
            
        self._cipher = cipher
        self._page_size = page_size
        
        self._request_semaphore = threading.Semaphore(max_connections)

        self._client_pages_index: dict[FeatureEntry, ClientPage] = {}
        self._client_pages: dict[int, ClientPage] = {}
        self._free_client_pages: set[ClientPage] = set()
        self._handler_map: dict[FeatureEntry, Callable[[UpdateContext], None]] = {}
        
        self._update_receiver_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._update_receiver_socket.bind((self._local_ip, client_port))
        self._update_receiver_port = self._update_receiver_socket.getsockname()[1]
        self._update_receiver_thread = threading.Thread(target=self._update_receiver, daemon=True)
        self._update_receiver_thread.start()
        
        self._client_registration_lock = threading.Lock()
        self._client_refresh_thread = threading.Thread(target=self._refresh_client_pages, daemon=True)
        self._client_refresh_thread.start()
        
    @property
    def clu_ip(self) -> str:
        return self._addr[0]

    @property
    def clu_port(self) -> int:
        return self._addr[1]

    @property
    def client_ip(self) -> str:
        return self._local_ip

    def send_request(self, msg: str, ignore_response: bool = False) -> str:
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.settimeout(self._timeout)
        
        payload = self._cipher.encrypt(msg.encode())
        
        try:
            with self._request_semaphore:
                sock.sendto(payload, self._addr)
                if not ignore_response:
                    resp, _ = sock.recvfrom(1024)
                    return self._cipher.decrypt(resp).decode()
                
        except Exception as e:
            raise e
        finally:
            sock.close()

    async def send_request_async(self, msg: str):
        return await asyncio.to_thread(self.send_request, msg)
    
    def check_alive(self) -> int:
        return int(self.send_lua_request("checkAlive()"), 16)
    
    async def check_alive_async(self) -> int:
        return await asyncio.to_thread(self.check_alive)

    def get_value(self, object_id: str, index: int):
        return self.send_lua_request(f"{object_id}:get({index})")

    async def get_value_async(self, object_id: str, index: int):
        return await asyncio.to_thread(self.get_value, object_id, index)

    def set_value(self, object_id: str, index: int, value) -> None:
        self.send_lua_request(f"{object_id}:set({index},{value})")

    async def set_value_async(self, object_id: str, index: int, value) -> None:
        await asyncio.to_thread(self.set_value, object_id, index, value)

    def execute_method(self, object_id, index, *args):
        args = list(map(lambda arg: f'"{arg}"' if isinstance(arg, str) else str(arg), args))
        args_str = ','.join([i for i in args]) if len(args) > 0 else '0'
        return self.send_lua_request(f"{object_id}:execute({index},{args_str})")
    
    async def execute_method_async(self, object_id, index, *args):
        return await asyncio.to_thread(self.execute_method, object_id, index, *args)
    
    # Client registration section
    def register_value_change_handler(self, object_id: str, index: int, handler: Callable[[UpdateContext], None]) -> None:
        with self._client_registration_lock:
            fentry = FeatureEntry(object_id, index)
            if fentry in self._handler_map.keys():
                self._handler_map[fentry] = handler
                return
            
            self._handler_map[fentry] = handler
            
            page: ClientPage = None
            if len(self._free_client_pages) > 0:
                page = self._free_client_pages.pop()
                if len(page.features) < self._page_size - 1:
                    self._free_client_pages.add(page)
                page.modified = True
                
            else:
                page = self._create_new_page()
                
            page.features.append(fentry)
            self._client_pages_index[fentry] = page
            
            self._refresh_page(page)
            
    async def register_value_change_handler_async(self, object_id: str, index: int, handler: Callable[[UpdateContext], None]) -> None:
        await asyncio.to_thread(self.register_value_change_handler, object_id, index, handler)
        
    def remove_value_change_handler(self, object_id: str, index: int) -> None:
        with self._client_registration_lock:
            fentry = FeatureEntry(object_id, index)
            del self._handler_map[fentry]
            
            page = self._client_pages_index.pop(fentry)
            page.features.remove(fentry)
            page.modified = True
            
            if len(page.features) == 0:
                self._free_client_pages.remove(page)
                self._client_pages.pop(page.client_id)
            else:
                self._free_client_pages.add(page)
                
            self._refresh_page(page)
            
    async def remove_value_change_handler_async(self, object_id: str, index: int) -> None:
        await asyncio.to_thread(self.remove_value_change_handler, object_id, index)
        
    def send_lua_request(self, payload: str, ignore_response: bool = False, ignore_type: bool = False) -> str:
        req_id = generate_id_hex()
        
        if not (ignore_type or ignore_response):
            # basically remote code execution
            payload = f'(load("result = {payload} return (type(result) .. \\\":\\\" .. tostring(result))")())'
            
        payload = f'req:{self._local_ip}:{req_id}:{payload}' 
    
        resp = self.send_request(payload, ignore_response)
        
        if ignore_response:
            return None
        
        resp = extract_payload(resp)
        
        if ignore_type:
            return resp
        
        i = resp.find(":")
        
        resp_type = resp[:i]
        value = resp[i+1:]
        
        if resp_type == "number":
            return float(value)
        elif resp_type == "string":
            return value
        elif resp_type == "boolean":
            return value == "true"
        else:
            return None
        
    async def send_lua_request_async(self, payload: str, ignore_response: bool = False, ignore_type: bool = False) -> str:
        return await asyncio.to_thread(self.send_lua_request, payload, ignore_response, ignore_type)
        
    def _refresh_client_pages(self):
        while True:
            with self._client_registration_lock:
                try:
                    for page in self._client_pages.values():
                        self._refresh_page(page)
                except Exception as e:
                    print(e)
            
            time.sleep(self._client_refresh_interval)
        
    def _update_receiver(self) -> None:
        
        while True:
            try:
                encrypted, _ = self._update_receiver_socket.recvfrom(1024)
                decrypted = self._cipher.decrypt(encrypted).decode("utf-8")
            except:
                continue
            
            msg_time = time.time()
            with self._client_registration_lock:
                self._handle_update_message(decrypted, msg_time)
            
    def _handle_update_message(self, message: str, message_timestamp: float):
        client_id, values = _parse_update_message(message)
        page = self._client_pages.get(client_id, None)
        if page == None:
            return
        
        if page.modified:
            page.states = values
            page.modified = False
            
            if page.last_mod > message_timestamp:
                return
            
            for entry, value in zip(page.features, values):
                self._procces_update(entry, value)
                
        else:
            for entry, _, new_value in filter(lambda x: x[1] != x[2], zip(page.features, page.states, values)):
                self._procces_update(entry, new_value)
            page.states = values
                
    def _procces_update(self, entry: FeatureEntry, value: Any):
        handler = self._handler_map[entry]
        threading.Thread(target=handler, args=(UpdateContext(entry.object_id, entry.index, value),)).start()
                
    def _create_new_page(self) -> ClientPage:
        used_ids = self._client_pages.keys()
        
        client_id = 1
        while client_id in used_ids:
            client_id += 1
            
        page = ClientPage(client_id)
        
        self._client_pages[client_id] = page
        self._free_client_pages.add(page)
        
        return page
    
    def _register_page(self, page: ClientPage) -> None:
        resp = self.send_lua_request(page.create_payload(self._local_ip, self._update_receiver_port), ignore_type=True)
        i = resp.find(':')
        self._handle_update_message(resp[i+1:], time.time())
    
    def _unregister_page(self, page: ClientPage) -> None:
        self.send_lua_request(f'SYSTEM:clientDestroy("{self._local_ip}",{self._update_receiver_port},{page.client_id})', ignore_response=True)
        
    def _refresh_page(self, page: ClientPage) -> None:
        self._unregister_page(page)
        self._register_page(page)
        page.last_mod = time.time()
        



import asyncio
import os
import threading
import time

import tftpy

from .cipher import GrentonCipher
from .clu_client import CluClient
from .exceptions import ConfigurationDownloadError, ConfigurationParserError
from .gobject import GObject
from .interface_manager import InterfaceManager
from .parsers.config_json_parser import parse_json
from .parsers.config_parser import parse_clu_config, CluConfig
from .parsers.om_parser import parse_om


def verify(ipaddress: str, key: str, iv: str) -> int | None:
    try:
        cipher = GrentonCipher(key, iv)
        clu_client = CluClient(ipaddress, 1234, cipher)
        
        return clu_client.check_alive()
    except:
        return None

async def verify_async(ipaddress: str, key: str, iv: str) -> int | None:
    return await asyncio.to_thread(verify, ipaddress, key, iv)

class GrentonApi:
    
    _clu_config: CluConfig
    
    _interface_manager: InterfaceManager = None
    _interface_manager_lock: threading.Lock = threading.Lock()
    
    def __init__(
        self,
        ipaddress,
        key: str | bytes,
        iv: str | bytes,
        cache_dir = "pygrenton_cache",
        timeout: float = 1,
        client_ip: str | None = None,
        client_port: int = 0,
        max_connections: int = 6
    ) -> None:
        self._ipaddress = ipaddress
        self._cache_dir = cache_dir
        if not os.path.exists(cache_dir):
            os.mkdir(cache_dir)
        
        self._cipher = GrentonCipher(key, iv)
        self._clu_client = CluClient(ipaddress, 1234, self._cipher, timeout, client_ip=client_ip, client_port=client_port, max_connections=max_connections)
        
        clu_sn = self._clu_client.check_alive()
        self._config_cache_dir = os.path.join(cache_dir, str(clu_sn))
        if not os.path.exists(self._config_cache_dir):
            os.mkdir(self._config_cache_dir)
        
        with self._interface_manager_lock:
            if self._interface_manager is None:
                GrentonApi._interface_manager = InterfaceManager(cache_dir)
        
        self._download_config()
        self._parse_config()
        
    @property
    def clu_config(self) -> CluConfig:
        return self._clu_config
        
    @property
    def objects(self) -> list[GObject]:
        return list(self._clu_config.objects_by_id.values())
    
    @property
    def objects_by_id(self) -> dict[str, GObject]:
        return self._clu_config.objects_by_id
    
    @property
    def objects_by_name(self) -> dict[str, GObject]:
        return self._clu_config.objects_by_name
        
    def get_objects_by_class(self, class_int: int) -> list[GObject]:
        return self._clu_config.objects_by_class.get(class_int, [])
    
    def get_object_by_id(self, object_id: str) -> GObject | None:
        return self._clu_config.objects_by_id.get(object_id, None)
    
    def get_object_by_name(self, name: str) -> GObject | None:
        return self._clu_config.objects_by_name.get(name, None)
        
    async def check_alive_async(self) -> int:
        return await self._clu_client.check_alive_async()
    
    def check_alive(self) -> int:
        return self._clu_client.check_alive()
        
    def register_update_handlers(self):
        self._clu_client.start_client_registration()
        
    def _download_config(self) -> None:
        
        # a trick to speed up the download by ignoring rest of the content of the file
        def skip_useless_part(data):
            msg =  data.data.decode()
            if msg.find("EventsFor") != -1:
                data.data = data.data[:-1]
        
        try:
            tftp_client = tftpy.TftpClient(self._ipaddress, 69, options={"tsize": 0})
            self._clu_client.send_request("req_start_ftp")
            time.sleep(0.01)
            tftp_client.download("a:\\CONFIG.JSON", os.path.join(self._config_cache_dir, "config.json"))
            tftp_client.download("a:\\om.lua", os.path.join(self._config_cache_dir, "om.lua"), packethook=skip_useless_part)
            self._clu_client.send_request("req_tftp_stop")
        except:
            raise ConfigurationDownloadError
        
    def _parse_config(self) -> None:
        try:
            config_json = None
            om = None
            with open(os.path.join(self._config_cache_dir, "config.json"), "r") as f:
                config_json = parse_json(f)
            with open(os.path.join(self._config_cache_dir, "om.lua"), "r") as f:
                om = parse_om(f)
                
            self._clu_config = parse_clu_config(config_json, om, self._interface_manager, self._clu_client)
        except:
            raise ConfigurationParserError
            
        


from base64 import b64decode

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes


class GrentonCipher:

    def __init__(self, key: str | bytes, iv: str | bytes) -> None:
        if isinstance(key, str):
            self._key = b64decode(key)
        else:
            self._key = key
            
        if isinstance(iv, str):
            self._iv = b64decode(iv)
        else:
            self._iv = iv
            
        self._padding = padding.PKCS7(128)
        self._cipher = Cipher(
            algorithms.AES(self._key), modes.CBC(self._iv), backend=default_backend()
        )

    def encrypt(self, data: bytes) -> bytes:
        encryptor = self._cipher.encryptor()
        padder = self._padding.padder()
        data = padder.update(data) + padder.finalize()
        return encryptor.update(data) + encryptor.finalize()

    def decrypt(self, data: bytes) -> bytes:
        decryptor = self._cipher.decryptor()
        unpadder = self._padding.unpadder()
        data = decryptor.update(data) + decryptor.finalize()
        return unpadder.update(data) + unpadder.finalize()




import pytest

from pygrenton.cipher import GrentonCipher

@pytest.fixture
def cipher() -> GrentonCipher:
    key = "RwHGl8yQVJmU8mc4s1KNKg=="
    iv = "Kv2bAZd7nJCsXUWCX+lbMA=="
    return GrentonCipher(key, iv)

def test_grenton_cipher_encrypt(cipher: GrentonCipher):
    encrypted_message = b'\xf8\xf5\xf8\xd5\xfc\xdb\x08O\xd1[\xa9\x99I\x1f\xa3^'
    assert cipher.encrypt("test_message".encode("utf-8")) == encrypted_message
    
def test_grenton_cipher_decrypt(cipher: GrentonCipher):
    encrypted_message = b'\xf8\xf5\xf8\xd5\xfc\xdb\x08O\xd1[\xa9\x99I\x1f\xa3^'
    assert cipher.decrypt(encrypted_message).decode("utf-8") == "test_message"



from dataclasses import dataclass, field

from .types import CallType, DataType, ModuleObjectType


@dataclass
class FeatureInterface:
    name: str
    index: int

    get: bool
    set: bool

    data_type: DataType
    unit: str

    enum: dict | None = None
    value_range: tuple[int, int] | None = None

@dataclass
class ParameterInterface:
    name: str
    data_type: DataType
    unit: str

    enum: list | None = None
    value_range: tuple[int, int] | None = None

@dataclass
class MethodInterface:
    name: str
    index: int
    call: CallType

    return_type: DataType
    unit: str | None

    parameters: list[ParameterInterface] = field(default_factory=list)

@dataclass
class CluObjectInterface:
    name: str
    obj_class: int
    version: int

    features: list[FeatureInterface] = field(default_factory=list)
    methods: list[MethodInterface] = field(default_factory=list)

@dataclass
class ModuleObjectInterface:
    name: str
    obj_class: int
    obj_type: ModuleObjectType

    features: list[FeatureInterface] = field(default_factory=list)
    methods: list[MethodInterface] = field(default_factory=list)

@dataclass
class ModuleInterface:
    name: str
    hw_type: int
    fw_type: int
    fw_api_version: int

    objects: dict[int, ModuleObjectInterface] = field(default_factory=dict)

@dataclass
class CluInterface:
    name: str
    hw_type: int
    hw_version: int
    fw_type: int
    fw_api_version: int

    features: list[FeatureInterface] = field(default_factory=list)
    methods: list[MethodInterface] = field(default_factory=list)

    objects: dict[int, CluObjectInterface] = field(default_factory=dict)



import os
import re
import xml.dom.minidom as md
from xml.dom.minidom import Element

from ..interfaces import *

VALUE_RANGE_TEMPLATE = re.compile(r"\d+\s*-\s*\d+")

def _parse_feature(elm: Element) -> FeatureInterface:
    name = elm.getAttribute("name")
    index = int(elm.getAttribute("index"))
    get = elm.getAttribute("get") == "true"
    set = elm.getAttribute("set") == "true"
    data_type = DataType(elm.getAttribute("type")) #TODO: make data type more generic
    unit = elm.getAttribute("unit")

    value_range = None
    vr = elm.getAttribute("range")
    if VALUE_RANGE_TEMPLATE.match(vr):
        value_range = tuple(int(x) for x in elm.getAttribute("range").split("-"))

    enum = None
    en = elm.getAttribute("enum")
    if en != "":
        if data_type == DataType.STRING:
            enum = {s: None for s in elm.getAttribute("enum").split(",")}
        elif data_type == DataType.NUMBER:
            enum = {int(s): None for s in elm.getAttribute("enum").split(",")}
            
        for node in elm.getElementsByTagName("enum"):
            name = node.getAttribute("name")
            value = node.getAttribute("value")
            enum[value] = name

    return FeatureInterface(name, index, get, set, data_type, unit, enum, value_range)

def _parse_parameter(elm: Element) -> ParameterInterface:
    name = elm.getAttribute("name")
    data_type = DataType(elm.getAttribute("type"))
    unit = elm.getAttribute("unit")

    value_range = None
    vr = elm.getAttribute("range")
    if VALUE_RANGE_TEMPLATE.match(vr):
        value_range = tuple(int(x) for x in elm.getAttribute("range").split("-"))

    enum = None
    en = elm.getAttribute("enum")
    if en != "":
        if data_type == DataType.STRING:
            enum = [s for s in elm.getAttribute("enum").split(",")]
        elif data_type == DataType.NUMBER:
            enum = [int(s) for s in elm.getAttribute("enum").split(",")]

    return ParameterInterface(name, data_type, unit, enum, value_range)

def _parse_method(elm: Element) -> MethodInterface:
    call = CallType(elm.getAttribute("call"))
    index = int(elm.getAttribute("index"))
    name = elm.getAttribute("name")

    # make return type more generic
    return_value = DataType(elm.getAttribute("return"))

    params = [_parse_parameter(par) for par in elm.getElementsByTagName("param")]

    return MethodInterface(name, index, call, return_value, None, params)

def _parse_module_object(xml: Element) -> ModuleObjectInterface:

    obj_class = int(xml.getAttribute("class"))
    name = xml.getAttribute("name")
    obj_type = ModuleObjectType(xml.getAttribute("type"))

    features = [_parse_feature(x) for x in xml.getElementsByTagName("feature")]
    methods = [_parse_method(x) for x in xml.getElementsByTagName("method")]

    return ModuleObjectInterface(name, obj_class, obj_type, features, methods)

def parse_clu_object_xml(file) -> CluObjectInterface:
    dom = md.parse(file)

    obj = dom.getElementsByTagName("object")[0]
    obj_class = int(obj.getAttribute("class"))
    name = obj.getAttribute("name")
    version = int(obj.getAttribute("version"))

    features = [_parse_feature(xml) for xml in dom.getElementsByTagName("feature")]
    methods = [_parse_method(xml) for xml in dom.getElementsByTagName("method")]

    return CluObjectInterface(name, obj_class, version, features, methods)

def parse_module_xml(file) -> ModuleInterface:
    dom = md.parse(file)

    mod = dom.getElementsByTagName("module")[0]
    name = mod.getAttribute("name")
    hw_type = int(mod.getAttribute("typeId"), 16)

    fw = dom.getElementsByTagName("firmware")[0]
    fw_type = int(fw.getAttribute("typeId"), 16)
    fw_api_version = int(fw.getAttribute("version"), 16)

    objects = {}

    for obj_xml in dom.getElementsByTagName("object"):
        obj = _parse_module_object(obj_xml)
        objects[obj.obj_class] = obj

    return ModuleInterface(name, hw_type, fw_type, fw_api_version, objects)

def parse_clu_xml(file, objects_repo: dict[str, list[CluObjectInterface]]) -> CluInterface:
    dom = md.parse(file)

    clu = dom.getElementsByTagName("CLU")[0]
    name = clu.getAttribute("typeName")
    hw_type = int(clu.getAttribute("hardwareType"), 16)
    hw_version = int(clu.getAttribute("hardwareVersion"), 16)
    fw_type = int(clu.getAttribute("firmwareType"), 16)
    fw_version = int(clu.getAttribute("firmwareVersion"), 16)

    features = [_parse_feature(x) for x in dom.getElementsByTagName("feature")]
    methods = [_parse_method(x) for x in dom.getElementsByTagName("method")]

    objects = {}
    for obj in dom.getElementsByTagName("object"):
        obj_name = obj.getAttribute("name")
        obj_version = int(obj.getAttribute("version"))
        
        for obj_int in objects_repo[obj_name]:
            if obj_int.version == obj_version:
                objects[obj_int.obj_class] = obj_int

    return CluInterface(name, hw_type, hw_version, fw_type, fw_version, features, methods, objects)

def _append_elm(dict: dict, key, val):
    if key in dict:
        dict[key].append(val)
    else:
        dict[key] = [val]

def parse_interfaces(dir):

    clus: dict[int, list[CluInterface]] = {}
    modules: dict[int, list[ModuleInterface]] = {}
    objects: dict[str, list[CluObjectInterface]] = {}

    for path in os.scandir(dir):
        if not path.is_file() or not path.name.startswith("object_"):
            continue
        
        try:
            with open(path, 'r', encoding="utf-8") as file:
                obj = parse_clu_object_xml(file)
                _append_elm(objects, obj.name, obj)
        except IOError:
            pass

    for path in os.scandir(dir):
        if path.is_file():
            try:
                with open(path, "r", encoding="utf-8") as file:
                    if path.name.startswith("clu_"):
                        clu = parse_clu_xml(file, objects)
                        _append_elm(clus, clu.hw_type, clu)

                    elif path.name.startswith("module_"):
                        mod = parse_module_xml(file)
                        _append_elm(modules, mod.hw_type, mod)

            except IOError:
                pass

    for _, v in clus.items():
        v.sort(key=lambda x: x.fw_api_version)

    for _, v in modules.items():
        v.sort(key=lambda x: x.fw_api_version)

    return clus, modules




class FeatureNotGettableError(Exception):
    
    def __init__(self, name) -> None:
        message = f"Feature \"{name}\" is not gettable."
        super().__init__(message)

class FeatureNotSettableError(Exception):
    
    def __init__(self, name) -> None:
        message = f"Feature \"{name}\" is not settable."
        super().__init__(message)
        
class ConfigurationDownloadError(Exception):
    
    def __init__(self) -> None:
        super().__init__("Unable to download configuration.")
        
class ConfigurationParserError(Exception):
    
    def __init__(self) -> None:
        super().__init__("Unable to parse configuration.")



import json

from ..config_objects import CLUConfig, ModuleConfig


def _parse_module(json_elm) -> ModuleConfig:
    sn = json_elm["sn"]
    hw_type = json_elm["hwType"]
    hw_version = json_elm["hwVer"]
    fw_type = json_elm["fwType"]
    fw_api_version = json_elm["fwApiVer"]
    fw_version = json_elm["fwVer"]
    status = json_elm["status"]

    return ModuleConfig(sn, hw_type, hw_version, fw_type, fw_api_version, fw_version, status)

def parse_json(file) -> CLUConfig:
    json_doc = json.load(file)
    
    serial_number = json_doc["sn"]
    mac = json_doc["mac"]
    hw_type = json_doc["hwType"]
    hw_version = json_doc["hwVer"]
    fw_type = json_doc["fwType"]
    fw_api_version = json_doc["fwApiVer"]
    fw_version = json_doc["fwVer"]
    status = json_doc["status"]

    modules = {}

    for elm in json_doc["tfbusDevices"]:
        mod = _parse_module(elm)
        modules[mod.serial_number] = mod

    for elm in json_doc["zwaveDevices"]:
        mod = _parse_module(elm)
        modules[mod.serial_number] = mod

    return CLUConfig(serial_number, mac, hw_type, hw_version, fw_type, fw_api_version, fw_version, status, modules)





from dataclasses import dataclass


@dataclass
class ModuleConfig:
    serial_number: int
    hw_type: int
    hw_version: int
    fw_type: int
    fw_api_version: int
    fw_version: int
    status: str

@dataclass
class CLUConfig:
    serial_number: int
    mac: str
    hw_type: int
    hw_version: int
    fw_type: int
    fw_api_version: int
    fw_version: int
    status: str

    modules: dict[int, ModuleConfig]




import asyncio

from .clu_client import CluClient
from .interfaces import MethodInterface, ParameterInterface
from .types import CallType, DataType


class GMethod:
    _clu_client: CluClient
    _object_id: str
    _name: str
    _index: int
    _call_type: CallType

    _params: list[ParameterInterface]

    _return_type: DataType
    _unit: str | None
    
    def __init__(self, clu_client: CluClient, object_id: str, interface: MethodInterface) -> None:
        self._clu_client = clu_client
        self._object_id = object_id
        self._name = interface.name
        self._index = interface.index
        self._call_type = interface.call
        self._params = interface.parameters
        self._return_type = interface.return_type
        self._unit = interface.unit

    @property
    def name(self) -> str:
        return self._name

    @property
    def parent(self) -> str:
        return self._object_id

    @property
    def index(self) -> int:
        return self._index

    @property
    def call_type(self) -> CallType:
        return self._call_type

    @property
    def parameters(self) -> list[ParameterInterface]:
        return self._params

    @property
    def return_type(self) -> DataType:
        return self._return_type

    @property
    def unit(self) -> str | None:
        return self._unit

    async def execute_method_async(self, *args):
        if len(args) != len(self._params):
            raise ValueError(f"Incorrect number of arguments: expected {len(self._params)}, provided {len(args)}")
        
        for arg, intr in zip(args, self._params):
            
            if intr.data_type == DataType.NUMBER:
                if not (isinstance(arg, int) or isinstance(arg, float)):
                    raise ValueError(f"\"{type(arg)}\" is incorrect type for argument \"{intr.name}\". Expected a number")
            elif intr.data_type == DataType.STRING and not isinstance(arg, str):
                raise ValueError(f"\"{type(arg)}\" is incorrect type for argument \"{intr.name}\". Expected a string")

        value = None
        if self._call_type == CallType.SET:
            return await self._clu_client.set_value_async(self._object_id, self._index, args[0])
        elif self._call_type == CallType.GET:
            value = await self._clu_client.get_value_async(self._object_id, self._index)
        else:
            value = await self._clu_client.execute_method_async(self._object_id, self._index, *args)
         
        return self.return_type.convert_value(value)

    def execute_method(self, *args):
        return asyncio.get_event_loop().run_until_complete(self.execute_method_async(*args))




import asyncio

from .clu_client import CluClient
from .gfeature import GFeature
from .gmethod import GMethod
from .interfaces import CluInterface, CluObjectInterface, ModuleObjectInterface


class GObject:
    _clu_client: CluClient
    _name: str
    _object_id: str
    _obj_class: int
    _obj_class_name: str = ""

    _features: list[GFeature]
    _methods: list[GMethod]
    
    def __init__(self, clu_client: CluClient, name: str, object_id: str, interface: CluInterface | CluObjectInterface | ModuleObjectInterface) -> None:
        self._clu_client = clu_client
        self._name = name
        self._object_id = object_id
        
        if isinstance(interface, CluInterface):
            self._obj_class = 0
            self._obj_class_name = interface.name
        else:
            self._obj_class = interface.obj_class
            self._obj_class_name = interface.name

        self._features = [GFeature(clu_client, object_id, fint) for fint in interface.features]
        self._methods = [GMethod(clu_client, object_id, mint) for mint in interface.methods]

    @property
    def clu_client(self) -> CluClient:
        return self._clu_client

    @property
    def name(self) -> str:
        return self._name

    @property
    def object_id(self) -> str:
        return self._object_id

    @property
    def object_class(self) -> int:
        return self._obj_class
    
    @property
    def object_class_name(self) -> str:
        return self._obj_class_name

    @property
    def features(self) -> list[GFeature]:
        return self._features

    @property
    def methods(self) -> list[GMethod]:
        return self._methods
    
    def has_feature(self, key: int | str) -> bool:
        if isinstance(key, int):
            return self.get_feature_by_index(key) is not None
        if isinstance(key, str):
            return self.get_feature_by_name(key) is not None
        
        raise TypeError("Key must either be an int or a string.")

    def get_feature_by_name(self, name: str) -> GFeature | None:
        for feature in self._features:
            if feature.name == name:
                return feature
            
        return None

    def get_feature_by_index(self, index: int) -> GFeature | None:
        for feature in self._features:
            if feature.index == index:
                return feature
            
        return None
    
    def has_method(self, key: int | str) -> bool:
        if isinstance(key, int):
            return self.get_method_by_index(key) is not None
        if isinstance(key, str):
            return self.get_method_by_name(key) is not None
        
        raise TypeError("Key must either be an int or a string.")

    def get_method_by_name(self, name: str) -> GMethod | None:
        for method in self._methods:
            if method.name == name:
                return method
            
        return None

    def get_method_by_index(self, index: int) -> GMethod | None:
        for method in self._methods:
            if method.index == index:
                return method
            
        return None
    
    async def get_value_async(self, index: int):
        return await self._clu_client.get_value_async(self._object_id, index)

    def get_value(self, index: int):
        return asyncio.get_event_loop().run_until_complete(self.get_value_async(index))
    
    async def set_value_async(self, index: int, value) -> None: 
        await self._clu_client.set_value_async(self._object_id, index, value)

    def set_value(self, index: int, value) -> None:
        asyncio.get_event_loop().run_until_complete(self.set_value_async(index, value))
    
    async def execute_method_async(self, index: int, *args):
        return await self._clu_client.execute_method_async(self._object_id, index, args)

    def execute_method(self, index: int, *args):
        return asyncio.get_event_loop().run_until_complete(self.execute_method_async(index, args))